@@ -26,6 +26,12 @@
 
 ;;; Code:
 
+(defvar require-features nil)
+
+(defadvice require (before ivy-tests-require-hook (feature &rest _) activate)
+  "Record the requires into `require-features'."
+  (push feature require-features))
+
 (require 'ert)
 (require 'colir)
 
@@ -37,30 +43,45 @@
 
 (message "%s" (emacs-version))
 
+(ert-deftest ivy--lazy-load-ffap--ffap-url-p ()
+  (should (not (memq 'ffap require-features)))
+  (should (not (fboundp 'ffap-url-p)))
+  (should (string= (ivy-ffap-url-p "https://foo.org")
+                   "https://foo.org"))
+  (should (memq 'ffap require-features))
+  (should (fboundp 'ffap-url-p)))
+
 (defvar ivy-expr nil
   "Holds a test expression to evaluate with `ivy-eval'.")
 
 (defvar ivy-result nil
   "Holds the eval result of `ivy-expr' by `ivy-eval'.")
 
+(defvar ivy-eval-dir nil
+  "Hold the `default-directory' value to be used by `ivy-eval'.
+Since `execute-kbd-macro' doesn't pick up a let-bound `default-directory'.")
+
 (defun ivy-eval ()
   "Evaluate `ivy-expr'."
   (interactive)
-  (setq ivy-result (eval ivy-expr)))
+  (let ((default-directory (or ivy-eval-dir default-directory)))
+    (setq ivy-result (eval ivy-expr))))
 
 (global-set-key (kbd "C-c e") 'ivy-eval)
 
+(defvar ivy-test-inhibit-message t)
+
 (cl-defun ivy-with (expr keys &key dir)
   "Evaluate EXPR followed by KEYS."
   (let ((ivy-expr expr)
-        (inhibit-message t)
+        (inhibit-message ivy-test-inhibit-message)
         (buf (current-buffer)))
     (save-window-excursion
       (unwind-protect
            (progn
-             ;; `execute-kbd-macro' doesn't pick up `default-directory'
              (when dir
-               (dired (expand-file-name dir (counsel-locate-git-root))))
+               (setq dir (expand-file-name dir (counsel-locate-git-root))))
+             (setq ivy-eval-dir dir)
              (execute-kbd-macro
               (vconcat (kbd "C-c e")
                        (kbd keys))))
@@ -181,7 +202,11 @@ will bring the behavior in line with the newer Emacsen."
   (should (string= (swiper--re-builder "^a")
                    "^ ?\\(a\\)"))
   (should (string= (swiper--re-builder "^a b")
-                   "^ \\(a\\).*?\\(b\\)")))
+                   "^ \\(a\\).*?\\(b\\)"))
+  (should
+   (string-match-p
+    "\\`\\\\_<.*\\\\_>\\'"
+    (swiper--re-builder "\\_<iv\\_>"))))
 
 (ert-deftest swiper--re-builder-char-fold ()
   :expected-result (if (>= emacs-major-version 25)
@@ -219,7 +244,9 @@ will bring the behavior in line with the newer Emacsen."
                  "foo\\["))
   (should (equal (ivy--regex
                   ".org")
-                 "\\.org")))
+                 "\\.org"))
+  (should (equal (ivy--regex "foo\\") "foo"))
+  (should (equal (ivy--regex "foo\\|") "foo")))
 
 (ert-deftest ivy--split-negation ()
   (should (equal (ivy--split-negation "") ()))
@@ -288,7 +315,9 @@ will bring the behavior in line with the newer Emacsen."
   (should (string= (ivy--regex-fuzzy "^")
                    "^"))
   (should (string= (ivy--regex-fuzzy "$")
-                   "$")))
+                   "$"))
+  (should (equal (ivy--regex-fuzzy "abc\\|")
+                 "\\(a\\)[^b\n]*\\(b\\)[^c\n]*\\(c\\)")))
 
 (ert-deftest ivy--regex-ignore-order ()
   (should (equal (ivy--regex-ignore-order "tmux")
@@ -320,7 +349,9 @@ will bring the behavior in line with the newer Emacsen."
   (ivy--regex-ignore-order "! ! !")
   ;; Escape invalid regexps.
   (should (equal (ivy--regex-ignore-order "foo[ bar[xy]")
-                 '(("foo\\[" . t) ("bar[xy]" . t)))))
+                 '(("foo\\[" . t) ("bar[xy]" . t))))
+  (should (equal (ivy--regex-ignore-order "foo\\|")
+                 '(("foo" . t)))))
 
 (ert-deftest ivy--format ()
   (should (string= (let ((ivy--index 10)
@@ -723,7 +754,7 @@ will bring the behavior in line with the newer Emacsen."
        (switch-to-buffer standard-output t)
        ,expr
        (ivy-mode)
-       (let ((inhibit-message t))
+       (let ((inhibit-message ivy-test-inhibit-message))
          (execute-kbd-macro
           ,(apply #'vconcat (mapcar #'kbd keys)))))))
 
@@ -919,11 +950,38 @@ will bring the behavior in line with the newer Emacsen."
     (should (eq (ivy--sort-function 'c) fn1))))
 
 (ert-deftest ivy-read-directory-name ()
+  (ivy-mode 1)
   (should
-   (equal "/tmp/"
+   (equal (expand-file-name "/tmp/")
           (ivy-with
            '(read-directory-name "cd: " "/tmp")
-           "RET"))))
+           "RET")))
+  (should
+   (equal (expand-file-name "/tmp")
+          (ivy-with
+           '(read-directory-name "cd: ")
+           "C-M-j"
+           :dir "/tmp")))
+  (should
+   (equal (expand-file-name "/tmp/")
+          (ivy-with
+           '(read-directory-name "cd: ")
+           "tmp C-j C-M-j"
+           :dir "/")))
+  (should
+   (equal (expand-file-name "/")
+          (ivy-with
+           '(read-directory-name "cd: ")
+           "DEL C-M-j"
+           :dir "/tmp"))))
+
+(ert-deftest ivy-counsel-read-directory-name ()
+  (should
+   (equal (expand-file-name "/tmp/")
+          (ivy-with
+           '(counsel-read-directory-name "cd: ")
+           "RET"
+           :dir "/tmp/"))))
 
 (ert-deftest ivy-partial-files ()
   (when (file-exists-p "/tmp/ivy-partial-test")
@@ -957,25 +1015,21 @@ will bring the behavior in line with the newer Emacsen."
              (kill-buffer temp-buffer))))))
 
 (ert-deftest counsel-yank-pop ()
-  (let ((kill-ring '("foo")))
-    (should (equal
-             (ivy-with-temp-buffer '(counsel-yank-pop) "C-m")
-             '(4 "foo")))
-    (let ((counsel-yank-pop-after-point t))
-      (should (equal
-               (ivy-with-temp-buffer '(counsel-yank-pop) "C-m")
-               '(1 "foo"))))))
+  (should (equal
+           (let ((kill-ring '("foo")))
+             (ivy-with-temp-buffer '(counsel-yank-pop) "C-m"))
+           '(4 "foo")))
+  (should (equal
+           (let ((kill-ring '("foo"))
+                 (counsel-yank-pop-after-point t))
+             (ivy-with-temp-buffer '(counsel-yank-pop) "C-m"))
+           '(1 "foo"))))
 
 (ert-deftest ivy-read-file-name-in-buffer-visiting-file ()
   "Test `ivy-immediate-done' command in `read-file-name' without any editing in
 a buffer visiting a file."
   (let ((ivy-mode-reset-arg (if ivy-mode 1 0)))
     (ivy-mode 1)
-    ;; `ivy-read' returns "~/dummy-dir/dummy-file" (same object, not a copy).
-    ;;
-    ;; `read-file-name-default' will then return "" in order for
-    ;; `set-visited-file-name' to detect that the user typed RET with
-    ;; the minibuffer empty.
     (should
      (equal (ivy-with
              '(let ((insert-default-directory t))
@@ -984,11 +1038,22 @@ a buffer visiting a file."
                  (read-file-name "Load file: " nil nil 'lambda)))
              ;; No editing, just command ivy-immediate-done
              "C-M-j")
-            ""))
+            "~/dummy-dir/dummy-file"))
     (should
      (equal (ivy-state-current ivy-last) "~/dummy-dir/dummy-file"))
     (ivy-mode ivy-mode-reset-arg)))
 
+(ert-deftest ivy-read-file-name-make-directory ()
+  (ivy-mode 1)
+  (should
+   (equal
+    (ivy-with
+     '(read-file-name "Make directory: " default-directory default-directory
+       nil nil)
+     "C-M-j"
+     :dir "/tmp/non-existant-dir/")
+    (expand-file-name "/tmp/non-existant-dir/"))))
+
 (ert-deftest ivy-starts-with-dotslash ()
   (should (ivy--starts-with-dotslash "./test1"))
   (should (ivy--starts-with-dotslash ".\\test2"))
@@ -1044,16 +1109,17 @@ a buffer visiting a file."
                         (let ((key (eval (cadr x))))
                           (list key (lookup-key global-map key)))))
                     body))))
-    `(let ((temp-buffer (generate-new-buffer " *temp*")))
+    `(let ((temp-buffer (get-buffer-create " *temp*")))
        (save-window-excursion
          (unwind-protect
               (progn
                 (switch-to-buffer temp-buffer)
+                (erase-buffer)
                 (insert ,text)
                 (search-backward "|")
                 (delete-char 1)
                 (setq current-prefix-arg nil)
-                (let ((inhibit-message t))
+                (let ((inhibit-message ivy-test-inhibit-message))
                   ,@(mapcar (lambda (x)
                               (if (and (listp x)
                                        (stringp (car x)))
@@ -1070,6 +1136,22 @@ a buffer visiting a file."
            (and (buffer-name temp-buffer)
                 (kill-buffer temp-buffer)))))))
 
+(ert-deftest swiper-thing-at-point ()
+  (should
+   (string=
+    (ivy-with-text
+     "let\n|let\nlet"
+     (global-set-key (kbd "C-s") #'swiper-thing-at-point)
+     ("C-s" "RET"))
+    "let\nlet|\nlet"))
+  (should
+   (string=
+    (ivy-with-text
+     "foo\nlet\nbar\n|let\nlet"
+     (global-set-key (kbd "C-s") #'swiper-thing-at-point)
+     ("C-s" "RET"))
+    "foo\nlet\nbar\nlet|\nlet")))
+
 (ert-deftest swiper-isearch ()
   (should
    (string=
@@ -1129,24 +1211,12 @@ a buffer visiting a file."
      (global-set-key (kbd "C-r") #'isearch-backward-regexp)
      ("C-r" "defun\\|defvar" "RET"))
     "(defun foo)\nasdf\n(|defvar bar)"))
-  ;; NOTE: The following two behaviors do not match
-  ;; `isearch-backward-regexp', but they match that of
-  ;; `swiper-isearch-forward', as `swiper-isearch' does not reset the
-  ;; point when the regexp becomes invalid, meaning the point is left
-  ;; at the initial match of the first part of the regexp.
   (should
    (string=
     (ivy-with-text
      "(defun foo)\nasdf\n(defvar bar)|"
      (global-set-key (kbd "C-r") #'swiper-isearch-backward)
      ("C-r" "defun\\|defvar" "RET"))
-    "(|defun foo)\nasdf\n(defvar bar)"))
-  (should
-   (string=
-    (ivy-with-text
-     "(defun foo)\nasdf\n(defvar bar)|"
-     (global-set-key (kbd "C-r") #'swiper-isearch-backward)
-     ("C-r" "defun\\|defvar" "C-n RET"))
     "(defun foo)\nasdf\n(|defvar bar)"))
   (should
    (string=
@@ -1161,7 +1231,21 @@ a buffer visiting a file."
      "(defun foo)\nasdf\n(|defun bar)"
      (global-set-key (kbd "C-r") #'swiper-isearch-backward)
      ("C-r" "defun" "RET"))
-    "(|defun foo)\nasdf\n(defun bar)")))
+    "(|defun foo)\nasdf\n(defun bar)"))
+  (should
+   (string=
+    (ivy-with-text
+     "(defun foo)\nasdf\n(de|fun bar)"
+     (global-set-key (kbd "C-r") #'swiper-isearch-backward)
+     ("C-r" "def" "RET"))
+    "(|defun foo)\nasdf\n(defun bar)"))
+  (should
+   (string=
+    (ivy-with-text
+     "(defun foo)\nasdf\n(de|fun bar)"
+     (global-set-key (kbd "C-r") #'swiper-isearch-backward)
+     ("C-r" "def?" "RET"))
+    "(defun foo)\nasdf\n(|defun bar)")))
 
 (ert-deftest swiper-isearch-backward-backspace ()
   (should
@@ -1204,10 +1288,31 @@ a buffer visiting a file."
        ("C-s" "Foo" "C-n RET")))
     "Foo\nfoo|\nFOO\n")))
 
+(ert-deftest ivy-swiper-wgrep ()
+  :expected-result (if (and (= emacs-major-version 24)
+                            (<= emacs-minor-version 3))
+                       ;; `wgrep' requires at least 24.5
+                       :failed
+                     :passed)
+  (dolist (search-cmd '(swiper swiper-isearch))
+    (should
+     (string=
+      (let ((default-directory "/tmp/"))
+        (ivy-with-text
+         "|a one\na two\na three"
+         (global-set-key (kbd "C-s") search-cmd)
+         ("C-s" "a" "C-c C-o" "C-x C-q" "C-e" "1" "C-n" "2" "C-n" "C-e" "3")))
+      "-*- mode:grep; default-directory: \"/tmp/\" -*-
+
+
+3 candidates:
+./ *temp*:1:a one1
+./ *temp*:2:a two2
+./ *temp*:3:a three3|
+"))))
+
 (ert-deftest swiper--isearch-format ()
-  (setq swiper--isearch-point-history
-        (list
-         (cons "" 1)))
+  (setq swiper--isearch-start-point 0)
   (with-temp-buffer
     (insert
      "line0\nline1\nline line\nline line\nline5")
@@ -1252,17 +1357,7 @@ a buffer visiting a file."
   (should (equal (ivy--minibuffer-index-bounds 10 11 10) '(2 11 8)))
   (should (equal (ivy--minibuffer-index-bounds 1 3 10) '(0 3 1))))
 
-(defun counsel--setup-test-files ()
-  (unless (file-exists-p "tests/")
-    (shell-command
-     "git clone -b test --single-branch https://github.com/abo-abo/swiper/ tests"))
-  (let ((default-directory (expand-file-name "tests/"))
-        (version "066ec1d"))
-    (shell-command
-     (format "git checkout %s || git fetch && git checkout %s" version version))))
-
 (ert-deftest counsel-find-file-with-dollars ()
-  (counsel--setup-test-files)
   (should (string=
            (file-relative-name
             (ivy-with '(counsel-find-file) "fo C-m"
@@ -1270,7 +1365,6 @@ a buffer visiting a file."
            "tests/find-file/files-with-dollar/foo$")))
 
 (ert-deftest counsel-find-file-with-dotfiles ()
-  (counsel--setup-test-files)
   (should (string=
            (file-relative-name
             (ivy-with '(counsel-find-file) "f C-m"
@@ -1283,7 +1377,6 @@ a buffer visiting a file."
            "tests/find-file/dotfiles/.foobar1")))
 
 (ert-deftest counsel-find-file-with-spaces ()
-  (counsel--setup-test-files)
   (let ((ivy-extra-directories nil))
     (should (string=
              (file-relative-name
@@ -1308,6 +1401,37 @@ a buffer visiting a file."
       (should (string= (ivy-with read-numbers "C-' a") "1"))
       (should (string= (ivy-with read-numbers "C-v C-' d") "7")))))
 
+(ert-deftest ivy--yank-handle-case-fold ()
+  (should (string=
+           (let ((ivy-text ""))
+             (ivy--yank-handle-case-fold "FirstName"))
+           "FirstName"))
+  (should (string=
+           (let ((ivy-text "f"))
+             (ivy--yank-handle-case-fold "irstName"))
+           "irstname")))
+
+(ert-deftest ivy--handle-directory ()
+  (should (string= (ivy--handle-directory "/") "/"))
+  (should (string= (let ((ivy--directory "/tmp/"))
+                     (ivy--handle-directory "/sudo::"))
+                   "/sudo::/tmp/")))
+
+(defun ivy-test-run-tests ()
+  (let ((test-sets
+         '(
+           ;; this test must run first as other tests might force a load
+           ivy--lazy-load-ffap--ffap-url-p
+           ;; run the rest of the tests
+           (not ivy--lazy-load-ffap--ffap-url-p)))
+        (unexpected 0))
+    (dolist (test-set test-sets)
+      (cl-incf
+       unexpected
+       (ert-stats-completed-unexpected
+        (ert-run-tests-batch test-set))))
+    (kill-emacs (if (zerop unexpected) 0 1))))
+
 (provide 'ivy-test)
 
 ;;; ivy-test.el ends here
